""" Base GNodeSettings """

from gwbase.config.g_node_settings import GNodeSettings
from gwbase.config.g_node_settings import SupervisorSettings


__all__ = [
    "GNodeSettings",
    "SupervisorSettings",
]
